using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  [Creatable(CreatableAttribute.Categories.DataAnalysisRegression, Priority = 95)]
  [Item("M5RegressionTree", "A M5 regression tree / rule set classifier")]
  public sealed class M5Regression : FixedDataAnalysisAlgorithm<IRegressionProblem> {
    #region Parametername
    private const string GenerateRulesParameterName = "GenerateRules";
    private const string ImpurityParameterName = "Split";
    private const string MinimalNodeSizeParameterName = "MinimalNodeSize";
    private const string ModelTypeParameterName = "ModelType";
    private const string PruningTypeParameterName = "PruningType";
    private const string SeedParameterName = "Seed";
    private const string SetSeedRandomlyParameterName = "SetSeedRandomly";
    #endregion

    #region Parameter properties
    public IFixedValueParameter<BoolValue> GenerateRulesParameter {
      get { return Parameters[GenerateRulesParameterName] as IFixedValueParameter<BoolValue>; }
    }
    public IConstrainedValueParameter<ISplitType> ImpurityParameter {
      get { return Parameters[ImpurityParameterName] as IConstrainedValueParameter<ISplitType>; }
    }
    public IFixedValueParameter<IntValue> MinimalNodeSizeParameter {
      get { return (IFixedValueParameter<IntValue>) Parameters[MinimalNodeSizeParameterName]; }
    }
    public IConstrainedValueParameter<ILeafType<IRegressionModel>> ModelTypeParameter {
      get { return Parameters[ModelTypeParameterName] as IConstrainedValueParameter<ILeafType<IRegressionModel>>; }
    }
    public IConstrainedValueParameter<IPruningType> PruningTypeParameter {
      get { return Parameters[PruningTypeParameterName] as IConstrainedValueParameter<IPruningType>; }
    }
    public IFixedValueParameter<IntValue> SeedParameter {
      get { return Parameters[SeedParameterName] as IFixedValueParameter<IntValue>; }
    }
    public IFixedValueParameter<BoolValue> SetSeedRandomlyParameter {
      get { return Parameters[SetSeedRandomlyParameterName] as IFixedValueParameter<BoolValue>; }
    }
    #endregion

    #region Properties 
    public bool GenerateRules {
      get { return GenerateRulesParameter.Value.Value; }
    }
    public ISplitType Split {
      get { return ImpurityParameter.Value; }
    }
    public int MinimalNodeSize {
      get { return MinimalNodeSizeParameter.Value.Value; }
    }
    public ILeafType<IRegressionModel> LeafType {
      get { return ModelTypeParameter.Value; }
    }
    public IPruningType PruningType {
      get { return PruningTypeParameter.Value; }
    }
    public int Seed {
      get { return SeedParameter.Value.Value; }
    }
    public bool SetSeedRandomly {
      get { return SetSeedRandomlyParameter.Value.Value; }
    }
    #endregion

    #region Constructors and Cloning
    [StorableConstructor]
    private M5Regression(bool deserializing) : base(deserializing) { }
    private M5Regression(M5Regression original, Cloner cloner) : base(original, cloner) { }
    public M5Regression() {
      var modelSet = new ItemSet<ILeafType<IRegressionModel>>(ApplicationManager.Manager.GetInstances<ILeafType<IRegressionModel>>());
      var pruningSet = new ItemSet<IPruningType>(ApplicationManager.Manager.GetInstances<IPruningType>());
      var impuritySet = new ItemSet<ISplitType>(ApplicationManager.Manager.GetInstances<ISplitType>());
      Parameters.Add(new FixedValueParameter<BoolValue>(GenerateRulesParameterName, "Whether a set of rules or a decision tree shall be created", new BoolValue(true)));
      Parameters.Add(new ConstrainedValueParameter<ISplitType>(ImpurityParameterName, "The type of split function used to create node splits", impuritySet, impuritySet.OfType<OrderSplitType>().First()));
      Parameters.Add(new FixedValueParameter<IntValue>(MinimalNodeSizeParameterName, "The minimal number of samples in a leaf node", new IntValue(1)));
      Parameters.Add(new ConstrainedValueParameter<ILeafType<IRegressionModel>>(ModelTypeParameterName, "The type of model used for the nodes", modelSet, modelSet.OfType<LinearLeaf>().First()));
      Parameters.Add(new ConstrainedValueParameter<IPruningType>(PruningTypeParameterName, "The type of pruning used", pruningSet, pruningSet.OfType<M5LeafPruning>().First()));
      Parameters.Add(new FixedValueParameter<IntValue>(SeedParameterName, "The random seed used to initialize the new pseudo random number generator.", new IntValue(0)));
      Parameters.Add(new FixedValueParameter<BoolValue>(SetSeedRandomlyParameterName, "True if the random seed should be set to a random value, otherwise false.", new BoolValue(true)));
      Problem = new RegressionProblem();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new M5Regression(this, cloner);
    }
    #endregion

    protected override void Run(CancellationToken cancellationToken) {
      var random = new MersenneTwister();
      if (SetSeedRandomly) SeedParameter.Value.Value = new System.Random().Next();
      random.Reset(Seed);
      var solution = CreateM5RegressionSolution(Problem.ProblemData, random, LeafType, Split, PruningType, cancellationToken, MinimalNodeSize, GenerateRules, Results);
      AnalyzeSolution(solution);
    }

    #region Static Interface
    public static IRegressionSolution CreateM5RegressionSolution(IRegressionProblemData problemData, IRandom random,
      ILeafType<IRegressionModel> leafType = null, ISplitType splitType = null, IPruningType pruningType = null,
      CancellationToken? cancellationToken = null, int minNumInstances = 4, bool generateRules = false, ResultCollection results = null) {
      //set default values
      if (leafType == null) leafType = new LinearLeaf();
      if (splitType == null) splitType = new OrderSplitType();
      if (cancellationToken == null) cancellationToken = CancellationToken.None;
      if (pruningType == null) pruningType = new M5LeafPruning();


      var doubleVars = new HashSet<string>(problemData.Dataset.DoubleVariables);
      var vars = problemData.AllowedInputVariables.Concat(new[] {problemData.TargetVariable}).ToArray();
      if (vars.Any(v => !doubleVars.Contains(v))) throw new NotSupportedException("M5 regression does not support non-double valued input or output features.");

      var values = vars.Select(v => problemData.Dataset.GetDoubleValues(v, problemData.TrainingIndices).ToArray()).ToArray();
      if (values.Any(v => v.Any(x => double.IsNaN(x) || double.IsInfinity(x))))
        throw new NotSupportedException("M5 regression does not support NaN or infinity values in the input dataset.");
      var trainingData = new Dataset(vars, values);
      var pd = new RegressionProblemData(trainingData, problemData.AllowedInputVariables, problemData.TargetVariable);
      pd.TrainingPartition.End = pd.TestPartition.Start = pd.TestPartition.End = pd.Dataset.Rows;
      pd.TrainingPartition.Start = 0;

      //create & build Model
      var m5Params = new M5CreationParameters(pruningType, minNumInstances, leafType, pd, random, splitType, results);

      IReadOnlyList<int> t, h;
      pruningType.GenerateHoldOutSet(problemData.TrainingIndices.ToArray(), random, out t, out h);

      if (generateRules) {
        IM5MetaModel model = M5RuleSetModel.CreateRuleModel(problemData.TargetVariable, m5Params);
        model.BuildClassifier(t, h, m5Params, cancellationToken.Value);
        return model.CreateRegressionSolution(problemData);
      }
      else {
        IM5MetaModel model = M5TreeModel.CreateTreeModel(problemData.TargetVariable, m5Params);
        model.BuildClassifier(t, h, m5Params, cancellationToken.Value);
        return model.CreateRegressionSolution(problemData);
      }
    }

    public static void UpdateM5Model(M5TreeModel model, IRegressionProblemData problemData, IRandom random,
      ILeafType<IRegressionModel> leafType = null, CancellationToken? cancellationToken = null) {
      UpdateM5Model(model as IM5MetaModel, problemData, random, leafType, cancellationToken);
    }

    public static void UpdateM5Model(M5RuleSetModel model, IRegressionProblemData problemData, IRandom random,
      ILeafType<IRegressionModel> leafType = null, CancellationToken? cancellationToken = null) {
      UpdateM5Model(model as IM5MetaModel, problemData, random, leafType, cancellationToken);
    }

    private static void UpdateM5Model(IM5MetaModel model, IRegressionProblemData problemData, IRandom random,
      ILeafType<IRegressionModel> leafType = null, CancellationToken? cancellationToken = null) {
      if (cancellationToken == null) cancellationToken = CancellationToken.None;
      var m5Params = new M5UpdateParameters(leafType, problemData, random);
      model.UpdateModel(problemData.TrainingIndices.ToList(), m5Params, cancellationToken.Value);
    }
    #endregion

    #region Helpers
    private void AnalyzeSolution(IRegressionSolution solution) {
      Results.Add(new Result("RegressionSolution", (IItem) solution.Clone()));

      Dictionary<string, int> frequencies;
      if (!GenerateRules) {
        Results.Add(M5Analyzer.CreateLeafDepthHistogram((M5TreeModel) solution.Model));
        frequencies = M5Analyzer.GetTreeVariableFrequences((M5TreeModel) solution.Model);
      }
      else {
        Results.Add(M5Analyzer.CreateRulesResult((M5RuleSetModel) solution.Model, Problem.ProblemData, "M5TreeResult", true));
        frequencies = M5Analyzer.GetRuleVariableFrequences((M5RuleSetModel) solution.Model);
        Results.Add(M5Analyzer.CreateCoverageDiagram((M5RuleSetModel) solution.Model, Problem.ProblemData));
      }

      //Variable frequencies
      var sum = frequencies.Values.Sum();
      sum = sum == 0 ? 1 : sum;
      var impactArray = new DoubleArray(frequencies.Select(i => (double) i.Value / sum).ToArray()) {
        ElementNames = frequencies.Select(i => i.Key)
      };
      Results.Add(new Result("Variable Frequences", "relative frequencies of variables in rules and tree nodes", impactArray));
    }
    #endregion
  }
}