﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  public class M5TreeModel : RegressionModel, IM5MetaModel {
    private const string NoCurrentLeafesResultName = "Number of current Leafs";
    #region Properties
    [Storable]
    internal M5NodeModel Root { get; private set; }
    //[Storable]
    //private M5Parameters M5Params { get; set; }
    #endregion

    #region HLConstructors & Cloning
    [StorableConstructor]
    protected M5TreeModel(bool deserializing) : base(deserializing) { }
    protected M5TreeModel(M5TreeModel original, Cloner cloner) : base(original, cloner) {
      Root = cloner.Clone(original.Root);
    }
    protected M5TreeModel(string targetVariable) : base(targetVariable) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new M5TreeModel(this, cloner);
    }
    #endregion

    internal static M5TreeModel CreateTreeModel(string targetAttr, M5CreationParameters m5CreationParams) {
      return m5CreationParams.LeafType is ILeafType<IConfidenceRegressionModel> ? new ConfidenceM5TreeModel(targetAttr) : new M5TreeModel(targetAttr);
    }

    #region RegressionModel
    public override IEnumerable<string> VariablesUsedForPrediction {
      get { return Root.VariablesUsedForPrediction ?? new List<string>(); }
    }
    public override IEnumerable<double> GetEstimatedValues(IDataset dataset, IEnumerable<int> rows) {
      if (Root == null) throw new NotSupportedException("The classifier has not been built yet");
      return Root.GetEstimatedValues(dataset, rows);
    }
    public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      return new RegressionSolution(this, problemData);
    }
    #endregion

    #region IM5Component
    void IM5MetaModel.BuildClassifier(IReadOnlyList<int> trainingRows, IReadOnlyList<int> holdoutRows, M5CreationParameters m5CreationParams, CancellationToken cancellation) {
      Root = null;
      var globalStdDev = m5CreationParams.Data.GetDoubleValues(m5CreationParams.TargetVariable, trainingRows).StandardDeviationPop();
      Root = M5NodeModel.CreateNode(m5CreationParams.TargetVariable, m5CreationParams);
      Root.Split(trainingRows, m5CreationParams, globalStdDev);
      InitializeLeafCounter(m5CreationParams);
      if (!(m5CreationParams.Pruningtype is NoPruning)) Root.Prune(trainingRows, holdoutRows, m5CreationParams, cancellation, globalStdDev);
      Root.InstallModels(trainingRows.Union(holdoutRows).ToArray(), m5CreationParams.Random, m5CreationParams.Data, m5CreationParams.LeafType, cancellation);
    }

    void IM5MetaModel.UpdateModel(IReadOnlyList<int> rows, M5UpdateParameters m5UpdateParameters, CancellationToken cancellation) {
      Root.InstallModels(rows, m5UpdateParameters.Random, m5UpdateParameters.Data, m5UpdateParameters.LeafType, cancellation);
    }
    #endregion

    #region Helpers
    private void InitializeLeafCounter(M5CreationParameters m5CreationParams) {
      if (!m5CreationParams.Results.ContainsKey(NoCurrentLeafesResultName))
        m5CreationParams.Results.Add(new Result(NoCurrentLeafesResultName, new IntValue(Root.EnumerateNodes().Count(x => x.IsLeaf))));
      else ((IntValue) m5CreationParams.Results[NoCurrentLeafesResultName].Value).Value = Root.EnumerateNodes().Count(x => x.IsLeaf);
    }
    #endregion

    [StorableClass]
    private class ConfidenceM5TreeModel : M5TreeModel, IConfidenceRegressionModel {
      #region HLConstructors & Cloning
      [StorableConstructor]
      protected ConfidenceM5TreeModel(bool deserializing) : base(deserializing) { }
      private ConfidenceM5TreeModel(ConfidenceM5TreeModel original, Cloner cloner) : base(original, cloner) { }
      public ConfidenceM5TreeModel(string targetVariable) : base(targetVariable) { }
      public override IDeepCloneable Clone(Cloner cloner) {
        return new ConfidenceM5TreeModel(this, cloner);
      }
      #endregion

      public IEnumerable<double> GetEstimatedVariances(IDataset dataset, IEnumerable<int> rows) {
        if (Root == null) throw new NotSupportedException("The classifier has not been built yet");
        return ((IConfidenceRegressionModel) Root).GetEstimatedVariances(dataset, rows);
      }
      public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
        return new ConfidenceRegressionSolution(this, problemData);
      }
    }
  }
}